import logging

from defi_services.abis.dex.pancakeswap.pancakeswap_v2_factory_abi import PANCAKESWAP_V2_FACTORY_ABI
from defi_services.abis.dex.spookyswap.masterchef_v2_abi import SPOOKYSWAP_MASTERCHEF_V2_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.dex_constant import Dex
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.dex.dex_info.spookyswap_info import SPOOKY_FTM_V2_INFO
from defi_services.services.dex.pancakeswap_v2_service import PancakeSwapV2Services

logger = logging.getLogger("Spooky V2 State Service")


class SpookySwapV2Info:
    mapping = {
        Chain.fantom: SPOOKY_FTM_V2_INFO
    }


class SpookySwapV2Services(PancakeSwapV2Services):
    def __init__(self, state_service: StateQuerier, chain_id: str = '0xfa'):
        super().__init__(state_service=state_service, chain_id=chain_id)

        self.pool_info = SpookySwapV2Info.mapping.get(chain_id)

        self.masterchef_abi = SPOOKYSWAP_MASTERCHEF_V2_ABI
        self.factory_abi = PANCAKESWAP_V2_FACTORY_ABI

    def get_service_info(self):
        info = {
            Dex.spooky_v2: {
                "chain_id": self.chain_id,
                "type": "dex",
                "protocol_info": self.pool_info
            }
        }
        return info

    # Get lp token info
    # def get_lp_token_function_info(self, supplied_data, block_number: int = "latest"):
    #     masterchef_addr = self.pool_info.get('master_chef_address')
    #
    #     rpc_calls = {}
    #
    #     lp_token_info = supplied_data['lp_token_info']
    #     for lp_token, info in lp_token_info.items():
    #         for fn_name in ["decimals", "totalSupply", "token0", "token1", "name"]:
    #             query_id = f"{fn_name}_{lp_token}_{block_number}".lower()
    #             rpc_calls[query_id] = self.state_service.get_function_info(
    #                 address=lp_token, abi=LP_TOKEN_ABI, fn_name=fn_name, fn_paras=None,
    #                 block_number=block_number)
    #
    #         query_id = f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower()
    #         rpc_calls[query_id] = self.state_service.get_function_info(
    #             address=lp_token, abi=LP_TOKEN_ABI, fn_name="balanceOf", fn_paras=[masterchef_addr],
    #             block_number=block_number)
    #
    #         if info.get('farming_pid') is not None:
    #             pid = int(info.get('farming_pid'))
    #
    #             query_id = f'getFarmData_{masterchef_addr}_{pid}_{block_number}'.lower()
    #             rpc_calls[query_id] = self.get_masterchef_function_info(
    #                 fn_name="getFarmData", fn_paras=[pid], block_number=block_number)
    #
    #     return rpc_calls

    # def decode_lp_token_info(self, supplied_data, decoded_data, block_number: int = "latest"):
    #     masterchef_addr = self.pool_info.get('master_chef_address')
    #
    #     result = {}
    #
    #     lp_token_info = supplied_data['lp_token_info']
    #     for lp_token, info in lp_token_info.items():
    #         token0 = decoded_data.get(f'token0_{lp_token}_{block_number}'.lower())
    #         token1 = decoded_data.get(f'token1_{lp_token}_{block_number}'.lower())
    #         decimals = decoded_data.get(f'decimals_{lp_token}_{block_number}'.lower())
    #         if (not token0) and (not token1) and (decimals is None):
    #             continue
    #
    #         total_supply = decoded_data.get(f'totalSupply_{lp_token}_{block_number}'.lower()) / 10 ** decimals
    #         name = decoded_data.get(f'name_{lp_token}_{block_number}'.lower())
    #
    #         staked_balance_query_id = f'balanceOf_{lp_token}_{masterchef_addr}_{block_number}'.lower()
    #         masterchef_balance = decoded_data.get(
    #             staked_balance_query_id) / 10 ** decimals
    #
    #         result[lp_token] = {
    #             "total_supply": total_supply,
    #             "token0": token0,
    #             "token1": token1,
    #             "name": name,
    #             'decimals': decimals,
    #             "stake_balance": masterchef_balance
    #         }
    #
    #         if info.get('farming_pid') is not None:
    #             pid = int(info.get('farming_pid'))
    #             farm_data_query_id = f'getFarmData_{masterchef_addr}_{pid}_{block_number}'.lower()
    #             farm_data = decoded_data.get(farm_data_query_id)
    #             acc_boo_per_share = farm_data[0][0] / 10 ** 18
    #             alloc_point = farm_data[0][2]
    #             total_alloc_point = farm_data[1]
    #             rewarder_addr = farm_data[2]
    #
    #             result[lp_token].update({
    #                 'acc_reward_per_share': acc_boo_per_share,
    #                 'alloc_point': alloc_point,
    #                 'total_alloc_point': total_alloc_point,
    #                 'rewarder_address': rewarder_addr,
    #                 'farming_pid': pid
    #             })
    #
    #     return result

    # User Reward
    def get_rewards_balance_function_info(self, wallet, supplied_data, block_number: int = "latest"):
        rpc_calls = {}

        reward_token = self.pool_info.get("reward_token")
        decimals_query_id = f'decimals_{reward_token}_{block_number}'.lower()
        rpc_calls[decimals_query_id] = self.state_service.get_function_info(
            address=reward_token, abi=ERC20_ABI, fn_name="decimals", block_number=block_number)

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if info.get('farming_pid') is not None:
                pid = int(info.get('farming_pid'))

                query_id = f'pendingBOO_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()
                rpc_calls[query_id] = self.get_masterchef_function_info(
                    fn_name="pendingBOO", fn_paras=[int(pid), wallet], block_number=block_number)

        return rpc_calls

    def calculate_rewards_balance(self, wallet: str, supplied_data: dict, decoded_data: dict,
                                  block_number: int = "latest") -> dict:
        reward_token = self.pool_info.get("reward_token")
        reward_decimals = decoded_data.get(f'decimals_{reward_token}_{block_number}'.lower())

        result = {}

        masterchef_addr = self.pool_info.get('master_chef_address')

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            if info.get('farming_pid') is not None:
                pid = int(info.get('farming_pid'))
                query_id = f'pendingBOO_{masterchef_addr}_{[pid, wallet]}_{block_number}'.lower()

                result[lp_token] = {reward_token: {'amount': decoded_data.get(query_id) / 10 ** reward_decimals}}

        return result



import json

LIDO_ALLOWED_RECIPIENTS_REGISTRY_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_admin",
        "type": "address"
      },
      {
        "internalType": "address[]",
        "name": "_addRecipientToAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_removeRecipientFromAllowedListRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_setParametersRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_updateSpentAmountRoleHolders",
        "type": "address[]"
      },
      {
        "internalType": "contract IBokkyPooBahsDateTimeContract",
        "name": "_bokkyPooBahsDateTimeContract",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_newAddress",
        "type": "address"
      }
    ],
    "name": "BokkyPooBahsDateTimeContractChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "uint256",
        "name": "_periodStartTimestamp",
        "type": "uint256"
      }
    ],
    "name": "CurrentPeriodAdvanced",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_limit",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_periodDurationMonths",
        "type": "uint256"
      }
    ],
    "name": "LimitsParametersChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "string",
        "name": "_title",
        "type": "string"
      }
    ],
    "name": "RecipientAdded",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      }
    ],
    "name": "RecipientRemoved",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "previousAdminRole",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "newAdminRole",
        "type": "bytes32"
      }
    ],
    "name": "RoleAdminChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleGranted",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleRevoked",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_alreadySpentAmount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_spendableBalance",
        "type": "uint256"
      },
      {
        "indexed": true,
        "internalType": "uint256",
        "name": "_periodStartTimestamp",
        "type": "uint256"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_periodEndTimestamp",
        "type": "uint256"
      }
    ],
    "name": "SpendableAmountChanged",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "_newSpentAmount",
        "type": "uint256"
      }
    ],
    "name": "SpentAmountChanged",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "ADD_RECIPIENT_TO_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "DEFAULT_ADMIN_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "REMOVE_RECIPIENT_FROM_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "SET_PARAMETERS_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "UPDATE_SPENT_AMOUNT_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "_title",
        "type": "string"
      }
    ],
    "name": "addRecipient",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "allowedRecipients",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "bokkyPooBahsDateTimeContract",
    "outputs": [
      {
        "internalType": "contract IBokkyPooBahsDateTimeContract",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAllowedRecipients",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getLimitParameters",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getPeriodState",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "_alreadySpentAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_spendableBalanceInPeriod",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodStartTimestamp",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodEndTimestamp",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      }
    ],
    "name": "getRoleAdmin",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "grantRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "hasRole",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      }
    ],
    "name": "isRecipientAllowed",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_payoutAmount",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_motionDuration",
        "type": "uint256"
      }
    ],
    "name": "isUnderSpendableBalance",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      }
    ],
    "name": "removeRecipient",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "renounceRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes32",
        "name": "role",
        "type": "bytes32"
      },
      {
        "internalType": "address",
        "name": "account",
        "type": "address"
      }
    ],
    "name": "revokeRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_bokkyPooBahsDateTimeContract",
        "type": "address"
      }
    ],
    "name": "setBokkyPooBahsDateTimeContract",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodDurationMonths",
        "type": "uint256"
      }
    ],
    "name": "setLimitParameters",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "spendableBalance",
    "outputs": [
      {
        "internalType": "uint256",
        "name": "",
        "type": "uint256"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes4",
        "name": "interfaceId",
        "type": "bytes4"
      }
    ],
    "name": "supportsInterface",
    "outputs": [
      {
        "internalType": "bool",
        "name": "",
        "type": "bool"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_newSpentAmount",
        "type": "uint256"
      }
    ],
    "name": "unsafeSetSpentAmount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_payoutAmount",
        "type": "uint256"
      }
    ],
    "name": "updateSpentAmount",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
''')


CHEF_INCENTIVES_CONTROLLER = [
    {
        "inputs": [
            {
                "internalType": "uint128[]",
                "name": "_startTimeOffset",
                "type": "uint128[]"
            },
            {
                "internalType": "uint128[]",
                "name": "_rewardsPerSecond",
                "type": "uint128[]"
            },
            {
                "internalType": "address",
                "name": "_poolConfigurator",
                "type": "address"
            },
            {
                "internalType": "contract IMultiFeeDistribution",
                "name": "_rewardMinter",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "_maxMintable",
                "type": "uint256"
            }
        ],
        "stateMutability": "nonpayable",
        "type": "constructor"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "token",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "user",
                "type": "address"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "balance",
                "type": "uint256"
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "totalSupply",
                "type": "uint256"
            }
        ],
        "name": "BalanceUpdated",
        "type": "event"
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "previousOwner",
                "type": "address"
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newOwner",
                "type": "address"
            }
        ],
        "name": "OwnershipTransferred",
        "type": "event"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_token",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "_allocPoint",
                "type": "uint256"
            }
        ],
        "name": "addPool",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address[]",
                "name": "_tokens",
                "type": "address[]"
            },
            {
                "internalType": "uint256[]",
                "name": "_allocPoints",
                "type": "uint256[]"
            }
        ],
        "name": "batchUpdateAllocPoint",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_user",
                "type": "address"
            },
            {
                "internalType": "address[]",
                "name": "_tokens",
                "type": "address[]"
            }
        ],
        "name": "claim",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "claimReceiver",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_user",
                "type": "address"
            },
            {
                "internalType": "address[]",
                "name": "_tokens",
                "type": "address[]"
            }
        ],
        "name": "claimableReward",
        "outputs": [
            {
                "internalType": "uint256[]",
                "name": "",
                "type": "uint256[]"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "name": "emissionSchedule",
        "outputs": [
            {
                "internalType": "uint128",
                "name": "startTimeOffset",
                "type": "uint128"
            },
            {
                "internalType": "uint128",
                "name": "rewardsPerSecond",
                "type": "uint128"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_user",
                "type": "address"
            },
            {
                "internalType": "uint256",
                "name": "_balance",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "_totalSupply",
                "type": "uint256"
            }
        ],
        "name": "handleAction",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "maxMintableTokens",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "mintedTokens",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "poolConfigurator",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "poolInfo",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "totalSupply",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "allocPoint",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "lastRewardTime",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "accRewardPerShare",
                "type": "uint256"
            },
            {
                "internalType": "contract IOnwardIncentivesController",
                "name": "onwardIncentives",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "poolLength",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "name": "registeredTokens",
        "outputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "renounceOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "rewardMinter",
        "outputs": [
            {
                "internalType": "contract IMultiFeeDistribution",
                "name": "",
                "type": "address"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "rewardsPerSecond",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_user",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "_receiver",
                "type": "address"
            }
        ],
        "name": "setClaimReceiver",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "_token",
                "type": "address"
            },
            {
                "internalType": "contract IOnwardIncentivesController",
                "name": "_incentives",
                "type": "address"
            }
        ],
        "name": "setOnwardIncentives",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "start",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "startTime",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [],
        "name": "totalAllocPoint",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "newOwner",
                "type": "address"
            }
        ],
        "name": "transferOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "userBaseClaimable",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    },
    {
        "inputs": [
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            },
            {
                "internalType": "address",
                "name": "",
                "type": "address"
            }
        ],
        "name": "userInfo",
        "outputs": [
            {
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256"
            },
            {
                "internalType": "uint256",
                "name": "rewardDebt",
                "type": "uint256"
            }
        ],
        "stateMutability": "view",
        "type": "function"
    }
]



import json

JUSTLEND_COMPTROLLER_ABI = json.loads('''
[
    {
      "inputs": [
        {
          "name": "action",
          "type": "string"
        },
        {
          "name": "pauseState",
          "type": "bool"
        }
      ],
      "name": "ActionPaused",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "action",
          "type": "string"
        },
        {
          "name": "pauseState",
          "type": "bool"
        }
      ],
      "name": "ActionPaused",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "ActionSupportedCollateralFactorGuardian",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "error",
          "type": "uint256"
        },
        {
          "name": "info",
          "type": "uint256"
        },
        {
          "name": "detail",
          "type": "uint256"
        }
      ],
      "name": "Failure",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "account",
          "type": "address"
        }
      ],
      "name": "MarketEntered",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "account",
          "type": "address"
        }
      ],
      "name": "MarketExited",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        }
      ],
      "name": "MarketListed",
      "type": "event"
    },
    {
      "inputs": [
        {
          "indexed": true,
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "newBorrowCap",
          "type": "uint256"
        }
      ],
      "name": "NewBorrowCap",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldBorrowCapGuardian",
          "type": "address"
        },
        {
          "name": "newBorrowCapGuardian",
          "type": "address"
        }
      ],
      "name": "NewBorrowCapGuardian",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldCloseFactorMantissa",
          "type": "uint256"
        },
        {
          "name": "newCloseFactorMantissa",
          "type": "uint256"
        }
      ],
      "name": "NewCloseFactor",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "oldCollateralFactorMantissa",
          "type": "uint256"
        },
        {
          "name": "newCollateralFactorMantissa",
          "type": "uint256"
        }
      ],
      "name": "NewCollateralFactor",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldCollateralFactorGuardian",
          "type": "address"
        },
        {
          "name": "newCollateralFactorGuardian",
          "type": "address"
        }
      ],
      "name": "NewCollateralFactorGuardian",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldLiquidationIncentiveMantissa",
          "type": "uint256"
        },
        {
          "name": "newLiquidationIncentiveMantissa",
          "type": "uint256"
        }
      ],
      "name": "NewLiquidationIncentive",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldMaxAssets",
          "type": "uint256"
        },
        {
          "name": "newMaxAssets",
          "type": "uint256"
        }
      ],
      "name": "NewMaxAssets",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldPauseGuardian",
          "type": "address"
        },
        {
          "name": "newPauseGuardian",
          "type": "address"
        }
      ],
      "name": "NewPauseGuardian",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "oldPriceOracle",
          "type": "address"
        },
        {
          "name": "newPriceOracle",
          "type": "address"
        }
      ],
      "name": "NewPriceOracle",
      "type": "event"
    },
    {
      "inputs": [
        {
          "name": "unitroller",
          "type": "address"
        }
      ],
      "name": "_become",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "name": "_borrowGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "name": "_mintGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "newBorrowCapGuardian",
          "type": "address"
        }
      ],
      "name": "_setBorrowCapGuardian",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "_setBorrowPaused",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "newCloseFactorMantissa",
          "type": "uint256"
        }
      ],
      "name": "_setCloseFactor",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "newCollateralFactorMantissa",
          "type": "uint256"
        }
      ],
      "name": "_setCollateralFactor",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "newCollateralFactorGuardian",
          "type": "address"
        }
      ],
      "name": "_setCollateralFactorGuardian",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "_setCollateralFactorGuardianStateForMarket",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "newLiquidationIncentiveMantissa",
          "type": "uint256"
        }
      ],
      "name": "_setLiquidationIncentive",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cTokens",
          "type": "address[]"
        },
        {
          "name": "newBorrowCaps",
          "type": "uint256[]"
        }
      ],
      "name": "_setMarketBorrowCaps",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "newMaxAssets",
          "type": "uint256"
        }
      ],
      "name": "_setMaxAssets",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "_setMintPaused",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "newPauseGuardian",
          "type": "address"
        }
      ],
      "name": "_setPauseGuardian",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "newOracle",
          "type": "address"
        }
      ],
      "name": "_setPriceOracle",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "inputs": [
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "_setSeizePaused",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "inputs": [
        {
          "name": "state",
          "type": "bool"
        }
      ],
      "name": "_setTransferPaused",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        }
      ],
      "name": "_supportMarket",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        },
        {
          "type": "uint256"
        }
      ],
      "name": "accountAssets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "admin",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "uint256"
        }
      ],
      "name": "allMarkets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "borrowAmount",
          "type": "uint256"
        }
      ],
      "name": "borrowAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "borrowCapGuardian",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "borrowCaps",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "borrowGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "borrowAmount",
          "type": "uint256"
        }
      ],
      "name": "borrowVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "name": "account",
          "type": "address"
        },
        {
          "name": "cToken",
          "type": "address"
        }
      ],
      "name": "checkMembership",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "name": "closeFactorMantissa",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "collateralFactorGuardian",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "collateralFactorGuardianMarkets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "compAccrued",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "name": "index",
          "type": "uint224"
        },
        {
          "name": "block",
          "type": "uint32"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "compBorrowState",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        },
        {
          "type": "address"
        }
      ],
      "name": "compBorrowerIndex",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "name": "compRate",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "compSpeeds",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        },
        {
          "type": "address"
        }
      ],
      "name": "compSupplierIndex",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "name": "index",
          "type": "uint224"
        },
        {
          "name": "block",
          "type": "uint32"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "compSupplyState",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "comptrollerImplementation",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "_cToken",
          "type": "address"
        }
      ],
      "name": "enterMarket",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256[]"
        }
      ],
      "inputs": [
        {
          "name": "cTokens",
          "type": "address[]"
        }
      ],
      "name": "enterMarkets",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cTokenAddress",
          "type": "address"
        }
      ],
      "name": "exitMarket",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        },
        {
          "type": "uint256"
        },
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "name": "account",
          "type": "address"
        }
      ],
      "name": "getAccountLiquidity",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [],
      "outputs": [
        {
          "type": "address[]"
        }
      ],
      "constant": true,
      "name": "getAllMarkets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address[]"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "name": "account",
          "type": "address"
        }
      ],
      "name": "getAssetsIn",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "name": "getBlockNumber",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "getCompAddress",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        },
        {
          "type": "uint256"
        },
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "name": "account",
          "type": "address"
        },
        {
          "name": "cTokenModify",
          "type": "address"
        },
        {
          "name": "redeemTokens",
          "type": "uint256"
        },
        {
          "name": "borrowAmount",
          "type": "uint256"
        }
      ],
      "name": "getHypotheticalAccountLiquidity",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "name": "isComptroller",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cTokenBorrowed",
          "type": "address"
        },
        {
          "name": "cTokenCollateral",
          "type": "address"
        },
        {
          "name": "liquidator",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "repayAmount",
          "type": "uint256"
        }
      ],
      "name": "liquidateBorrowAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cTokenBorrowed",
          "type": "address"
        },
        {
          "name": "cTokenCollateral",
          "type": "address"
        },
        {
          "name": "liquidator",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "actualRepayAmount",
          "type": "uint256"
        },
        {
          "name": "seizeTokens",
          "type": "uint256"
        }
      ],
      "name": "liquidateBorrowVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        },
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "name": "cTokenBorrowed",
          "type": "address"
        },
        {
          "name": "cTokenCollateral",
          "type": "address"
        },
        {
          "name": "actualRepayAmount",
          "type": "uint256"
        }
      ],
      "name": "liquidateCalculateSeizeTokens",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "name": "liquidationIncentiveMantissa",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "name": "isListed",
          "type": "bool"
        },
        {
          "name": "collateralFactorMantissa",
          "type": "uint256"
        },
        {
          "name": "isComped",
          "type": "bool"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "markets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "constant": true,
      "name": "maxAssets",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "minter",
          "type": "address"
        },
        {
          "name": "mintAmount",
          "type": "uint256"
        }
      ],
      "name": "mintAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "inputs": [
        {
          "type": "address"
        }
      ],
      "name": "mintGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "minter",
          "type": "address"
        },
        {
          "name": "actualMintAmount",
          "type": "uint256"
        },
        {
          "name": "mintTokens",
          "type": "uint256"
        }
      ],
      "name": "mintVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "oracle",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "pauseGuardian",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "pendingAdmin",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "address"
        }
      ],
      "constant": true,
      "name": "pendingComptrollerImplementation",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "redeemer",
          "type": "address"
        },
        {
          "name": "redeemTokens",
          "type": "uint256"
        }
      ],
      "name": "redeemAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "redeemer",
          "type": "address"
        },
        {
          "name": "redeemAmount",
          "type": "uint256"
        },
        {
          "name": "redeemTokens",
          "type": "uint256"
        }
      ],
      "name": "redeemVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "payer",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "repayAmount",
          "type": "uint256"
        }
      ],
      "name": "repayBorrowAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "payer",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "actualRepayAmount",
          "type": "uint256"
        },
        {
          "name": "borrowerIndex",
          "type": "uint256"
        }
      ],
      "name": "repayBorrowVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cTokenCollateral",
          "type": "address"
        },
        {
          "name": "cTokenBorrowed",
          "type": "address"
        },
        {
          "name": "liquidator",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "seizeTokens",
          "type": "uint256"
        }
      ],
      "name": "seizeAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "name": "seizeGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cTokenCollateral",
          "type": "address"
        },
        {
          "name": "cTokenBorrowed",
          "type": "address"
        },
        {
          "name": "liquidator",
          "type": "address"
        },
        {
          "name": "borrower",
          "type": "address"
        },
        {
          "name": "seizeTokens",
          "type": "uint256"
        }
      ],
      "name": "seizeVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "uint256"
        }
      ],
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "src",
          "type": "address"
        },
        {
          "name": "dst",
          "type": "address"
        },
        {
          "name": "transferTokens",
          "type": "uint256"
        }
      ],
      "name": "transferAllowed",
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "outputs": [
        {
          "type": "bool"
        }
      ],
      "constant": true,
      "name": "transferGuardianPaused",
      "stateMutability": "view",
      "type": "function"
    },
    {
      "inputs": [
        {
          "name": "cToken",
          "type": "address"
        },
        {
          "name": "src",
          "type": "address"
        },
        {
          "name": "dst",
          "type": "address"
        },
        {
          "name": "transferTokens",
          "type": "uint256"
        }
      ],
      "name": "transferVerify",
      "stateMutability": "nonpayable",
      "type": "function"
    }
  ]
''')


import json

LIDO_ALLOWED_RECIPIENTS_BUILDER_MULTI_TOKEN_TOOLING_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "contract IAllowedRecipientsFactory",
        "name": "_factory",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_admin",
        "type": "address"
      },
      {
        "internalType": "contract IEasyTrack",
        "name": "_easytrack",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_finance",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_bokkyPooBahsDateTimeContract",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "ADD_RECIPIENT_TO_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "ADD_TOKEN_TO_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "DEFAULT_ADMIN_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "REMOVE_RECIPIENT_FROM_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "REMOVE_TOKEN_FROM_ALLOWED_LIST_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "SET_PARAMETERS_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "UPDATE_SPENT_AMOUNT_ROLE",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "",
        "type": "bytes32"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "admin",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "bokkyPooBahsDateTimeContract",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "deployAddAllowedRecipient",
    "outputs": [
      {
        "internalType": "contract IAddAllowedRecipient",
        "name": "addAllowedRecipient",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "uint256",
        "name": "_limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodDurationMonths",
        "type": "uint256"
      },
      {
        "internalType": "address[]",
        "name": "_recipients",
        "type": "address[]"
      },
      {
        "internalType": "string[]",
        "name": "_titles",
        "type": "string[]"
      },
      {
        "internalType": "uint256",
        "name": "_spentAmount",
        "type": "uint256"
      },
      {
        "internalType": "bool",
        "name": "_grantRightsToEVMScriptExecutor",
        "type": "bool"
      }
    ],
    "name": "deployAllowedRecipientsRegistry",
    "outputs": [
      {
        "internalType": "contract IAllowedRecipientsRegistry",
        "name": "registry",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      }
    ],
    "name": "deployAllowedTokensRegistry",
    "outputs": [
      {
        "internalType": "contract IAllowedTokensRegistry",
        "name": "registry",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "_limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodDurationMonths",
        "type": "uint256"
      },
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      },
      {
        "internalType": "address[]",
        "name": "_recipients",
        "type": "address[]"
      },
      {
        "internalType": "string[]",
        "name": "_titles",
        "type": "string[]"
      },
      {
        "internalType": "uint256",
        "name": "_spentAmount",
        "type": "uint256"
      }
    ],
    "name": "deployFullSetup",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      }
    ],
    "name": "deployRemoveAllowedRecipient",
    "outputs": [
      {
        "internalType": "contract IRemoveAllowedRecipient",
        "name": "removeAllowedRecipient",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_recipient",
        "type": "address"
      },
      {
        "internalType": "string",
        "name": "_title",
        "type": "string"
      },
      {
        "internalType": "address[]",
        "name": "_tokens",
        "type": "address[]"
      },
      {
        "internalType": "uint256",
        "name": "_limit",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_periodDurationMonths",
        "type": "uint256"
      },
      {
        "internalType": "uint256",
        "name": "_spentAmount",
        "type": "uint256"
      }
    ],
    "name": "deploySingleRecipientTopUpOnlySetup",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedTokensRegistry",
        "type": "address"
      }
    ],
    "name": "deployTopUpAllowedRecipients",
    "outputs": [
      {
        "internalType": "contract ITopUpAllowedRecipients",
        "name": "topUpAllowedRecipients",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "easyTrack",
    "outputs": [
      {
        "internalType": "contract IEasyTrack",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "evmScriptExecutor",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "factory",
    "outputs": [
      {
        "internalType": "contract IAllowedRecipientsFactory",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "finance",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')



import json

MONEYMARKET = json.loads("""
[
  {
    "constant": true,
    "inputs": [],
    "name": "pendingAdmin",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "paused",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "oracle",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "liquidationDiscount",
    "outputs": [
      {
        "name": "mantissa",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "name": "markets",
    "outputs": [
      {
        "name": "isSupported",
        "type": "bool"
      },
      {
        "name": "blockNumber",
        "type": "uint256"
      },
      {
        "name": "interestRateModel",
        "type": "address"
      },
      {
        "name": "totalSupply",
        "type": "uint256"
      },
      {
        "name": "supplyRateMantissa",
        "type": "uint256"
      },
      {
        "name": "supplyIndex",
        "type": "uint256"
      },
      {
        "name": "totalBorrows",
        "type": "uint256"
      },
      {
        "name": "borrowRateMantissa",
        "type": "uint256"
      },
      {
        "name": "borrowIndex",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "collateralRatio",
    "outputs": [
      {
        "name": "mantissa",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "",
        "type": "address"
      },
      {
        "name": "",
        "type": "address"
      }
    ],
    "name": "supplyBalances",
    "outputs": [
      {
        "name": "principal",
        "type": "uint256"
      },
      {
        "name": "interestIndex",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "originationFee",
    "outputs": [
      {
        "name": "mantissa",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "name": "collateralMarkets",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "admin",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "",
        "type": "address"
      },
      {
        "name": "",
        "type": "address"
      }
    ],
    "name": "borrowBalances",
    "outputs": [
      {
        "name": "principal",
        "type": "uint256"
      },
      {
        "name": "interestIndex",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "payable": true,
    "stateMutability": "payable",
    "type": "fallback"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "account",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "startingBalance",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newBalance",
        "type": "uint256"
      }
    ],
    "name": "SupplyReceived",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "account",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "startingBalance",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newBalance",
        "type": "uint256"
      }
    ],
    "name": "SupplyWithdrawn",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "account",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "startingBalance",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "borrowAmountWithFee",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newBalance",
        "type": "uint256"
      }
    ],
    "name": "BorrowTaken",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "account",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "startingBalance",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newBalance",
        "type": "uint256"
      }
    ],
    "name": "BorrowRepaid",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "targetAccount",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "assetBorrow",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "borrowBalanceBefore",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "borrowBalanceAccumulated",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "amountRepaid",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "borrowBalanceAfter",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "liquidator",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "assetCollateral",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "collateralBalanceBefore",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "collateralBalanceAccumulated",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "amountSeized",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "collateralBalanceAfter",
        "type": "uint256"
      }
    ],
    "name": "BorrowLiquidated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldPendingAdmin",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newPendingAdmin",
        "type": "address"
      }
    ],
    "name": "NewPendingAdmin",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldAdmin",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newAdmin",
        "type": "address"
      }
    ],
    "name": "NewAdmin",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldOracle",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newOracle",
        "type": "address"
      }
    ],
    "name": "NewOracle",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "interestRateModel",
        "type": "address"
      }
    ],
    "name": "SupportedMarket",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldCollateralRatioMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newCollateralRatioMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "oldLiquidationDiscountMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newLiquidationDiscountMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewRiskParameters",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldOriginationFeeMantissa",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "newOriginationFeeMantissa",
        "type": "uint256"
      }
    ],
    "name": "NewOriginationFee",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "interestRateModel",
        "type": "address"
      }
    ],
    "name": "SetMarketInterestRateModel",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "equityAvailableBefore",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "amount",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "owner",
        "type": "address"
      }
    ],
    "name": "EquityWithdrawn",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "SuspendedMarket",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "newState",
        "type": "bool"
      }
    ],
    "name": "SetPaused",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "error",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "info",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "detail",
        "type": "uint256"
      }
    ],
    "name": "Failure",
    "type": "event"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "getCollateralMarketsLength",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "assetPrices",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "newPendingAdmin",
        "type": "address"
      }
    ],
    "name": "_setPendingAdmin",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [],
    "name": "_acceptAdmin",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "newOracle",
        "type": "address"
      }
    ],
    "name": "_setOracle",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "requestedState",
        "type": "bool"
      }
    ],
    "name": "_setPaused",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "account",
        "type": "address"
      }
    ],
    "name": "getAccountLiquidity",
    "outputs": [
      {
        "name": "",
        "type": "int256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "account",
        "type": "address"
      },
      {
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getSupplyBalance",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "account",
        "type": "address"
      },
      {
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "getBorrowBalance",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "interestRateModel",
        "type": "address"
      }
    ],
    "name": "_supportMarket",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      }
    ],
    "name": "_suspendMarket",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "collateralRatioMantissa",
        "type": "uint256"
      },
      {
        "name": "liquidationDiscountMantissa",
        "type": "uint256"
      }
    ],
    "name": "_setRiskParameters",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "originationFeeMantissa",
        "type": "uint256"
      }
    ],
    "name": "_setOriginationFee",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "interestRateModel",
        "type": "address"
      }
    ],
    "name": "_setMarketInterestRateModel",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "_withdrawEquity",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "supply",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "requestedAmount",
        "type": "uint256"
      }
    ],
    "name": "withdraw",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "userAddress",
        "type": "address"
      }
    ],
    "name": "calculateAccountValues",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      },
      {
        "name": "",
        "type": "uint256"
      },
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "repayBorrow",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "targetAccount",
        "type": "address"
      },
      {
        "name": "assetBorrow",
        "type": "address"
      },
      {
        "name": "assetCollateral",
        "type": "address"
      },
      {
        "name": "requestedAmountClose",
        "type": "uint256"
      }
    ],
    "name": "liquidateBorrow",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "asset",
        "type": "address"
      },
      {
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "borrow",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
""")
        


import logging

from defi_services.abis.dex.pancakeswap.pancakeswap_lp_token_abi import LP_TOKEN_ABI
from defi_services.abis.token.erc20_abi import ERC20_ABI
from defi_services.constants.chain_constant import Chain
from defi_services.constants.entities.dex_constant import Dex
from defi_services.jobs.queriers.state_querier import StateQuerier
from defi_services.services.dex.dex_info.uniswap_info import UNISWAP_ETH_INFO
from defi_services.services.dex_protocol_services import DexProtocolServices

logger = logging.getLogger("UniSwap V2 State Service")


class UniswapV2Info:
    mapping = {
        Chain.ethereum: UNISWAP_ETH_INFO
    }


class UniswapV2Services(DexProtocolServices):
    def __init__(self, state_service: StateQuerier, chain_id: str = '0x1'):
        super().__init__()
        self.chain_id = chain_id
        self.state_service = state_service
        self.pool_info = UniswapV2Info.mapping.get(chain_id)
        if self.pool_info:
            self.factory_abi = self.pool_info['factory_abi']

    def get_service_info(self):
        info = {
            Dex.uniswap_v2: {
                "chain_id": self.chain_id,
                "type": "dex",
                "protocol_info": self.pool_info
            }
        }
        return info

    def get_all_supported_lp_token(self, limit=10000, supplied_data: dict = None):
        web3 = self.state_service.get_w3()
        factory_addr = self.pool_info.get('factory_address')

        factory_contract = web3.eth.contract(
            address=web3.to_checksum_address(factory_addr), abi=self.factory_abi)
        pool_length = factory_contract.functions.allPairsLength().call()

        rpc_calls = {}
        for pid in range(0, min(pool_length, limit)):
            query_id = f'allPairs_{factory_addr}_{pid}_latest'.lower()
            rpc_calls[query_id] = self.get_factory_function_info(fn_name="allPairs", fn_paras=[pid])

        return rpc_calls

    def decode_all_supported_lp_token(self, limit: int = 10, decoded_data: dict = None,
                                      supplied_data: dict = None):
        result = {}
        for query_id, value in decoded_data.items():
            # Format query_id: f'allPairs_{self.factory_addr}_{pid}_latest'

            pid = int(query_id.split("_")[-2])
            result[value.lower()] = {"pid": pid}

        return result

    def get_farming_supported_lp_token(self, limit: int = 10) -> dict:
        return {}

    def decode_farming_supported_lp_token(self, response_data) -> dict:
        return {}

    def get_important_lp_token(self, supplied_data, block_number: int = "latest"):
        """Deprecated"""
        rpc_calls = {}
        lp_token_list = supplied_data['lp_token_list']
        for lp_token, value in lp_token_list.items():
            for fn_name in ["token0", "token1"]:
                query_id = f"{fn_name}_{lp_token}_{block_number}".lower()
                rpc_calls[query_id] = self.state_service.get_function_info(
                    address=lp_token, abi=LP_TOKEN_ABI, fn_name=fn_name, fn_paras=None,
                    block_number=block_number)

        return rpc_calls

    def decode_important_lp_token(self, supplied_data, response_data, block_number: int = "latest"):
        """Deprecated"""
        lp_token_list = supplied_data['lp_token_list']
        token_info = supplied_data['token_info']
        result = {}
        for lp_token, value in lp_token_list.items():
            pid = value.get('pid')
            token0 = response_data.get(f'token0_{lp_token}_{block_number}'.lower(), "")
            token1 = response_data.get(f'token1_{lp_token}_{block_number}'.lower(), "")
            # chỉ lấy những token mà có dữ liệu lưu trong database và có gía
            if token0 in token_info and token1 in token_info:
                result[lp_token] = {
                    'pid': pid,
                    "token0": token0,
                    "token1": token1
                }

        return result

    def get_lp_token_function_info(self, supplied_data, block_number: int = "latest"):
        rpc_calls = {}

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            for fn_name in ["decimals", "totalSupply", "token0", "token1", "name"]:
                query_id = f"{fn_name}_{lp_token}_{block_number}".lower()
                rpc_calls[query_id] = self.state_service.get_function_info(
                    address=lp_token, abi=LP_TOKEN_ABI, fn_name=fn_name, fn_paras=None,
                    block_number=block_number)

        return rpc_calls

    def decode_lp_token_info(self, supplied_data, decoded_data, block_number: int = "latest"):

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            try:
                token0 = decoded_data.get(f'token0_{lp_token}_{block_number}'.lower())
                token1 = decoded_data.get(f'token1_{lp_token}_{block_number}'.lower())
                decimals = decoded_data.get(f'decimals_{lp_token}_{block_number}'.lower())
                if (not token0) and (not token1) and (decimals is None):
                    continue

                total_supply = decoded_data.get(f'totalSupply_{lp_token}_{block_number}'.lower()) / 10 ** decimals
                name = decoded_data.get(f'name_{lp_token}_{block_number}'.lower())

                lp_token_info[lp_token].update({
                    "total_supply": total_supply,
                    "token0": token0,
                    "token1": token1,
                    "name": name,
                    'decimals': decimals
                })
            except Exception:
                continue
        return lp_token_info

    # Get balance of token
    def get_balance_of_token_function_info(self, supplied_data, block_number: int = "latest"):
        rpc_calls = {}

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            for token_key in ["token0", "token1"]:
                token_address = info.get(token_key, None)
                if token_address is not None:
                    balance_query_id = f'balanceOf_{token_address}_{lp_token}_{block_number}'.lower()
                    decimals_query_id = f'decimals_{token_address}_{block_number}'.lower()

                    rpc_calls[balance_query_id] = self.state_service.get_function_info(
                        address=token_address, abi=ERC20_ABI, fn_name="balanceOf", fn_paras=[lp_token],
                        block_number=block_number)

                    rpc_calls[decimals_query_id] = self.state_service.get_function_info(
                        address=token_address, abi=ERC20_ABI, fn_name="decimals", block_number=block_number)

        return rpc_calls

    def decode_balance_of_token_function_info(
            self, supplied_data, decoded_data, block_number: int = "latest"):

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            try:
                for token_key in ["token0", "token1"]:
                    token_address = info.get(token_key, None)
                    query_id = f'balanceOf_{token_address}_{lp_token}_{block_number}'.lower()

                    if (token_address is not None) and (decoded_data.get(query_id) is not None):
                        token_decimals = decoded_data.get(f'decimals_{token_address}_{block_number}'.lower()) or 18
                        lp_token_info[lp_token][f'{token_key}_amount'] = decoded_data.get(
                            query_id) / 10 ** token_decimals
            except Exception:
                continue
        return lp_token_info

    # User
    def get_user_info_function(self, wallet: str, supplied_data, stake: bool = False, block_number: int = "latest"):
        rpc_calls = {}

        lp_token_info = supplied_data['lp_token_info']
        # lượng token đang hold trong ví
        for lp_token, value in lp_token_info.items():
            query_id = f'balanceOf_{lp_token}_{wallet}_{block_number}'.lower()
            rpc_calls[query_id] = self.state_service.get_function_info(
                address=lp_token, abi=LP_TOKEN_ABI, fn_name="balanceOf",
                fn_paras=[wallet], block_number=block_number)

            decimals_query_id = f'decimals_{lp_token}_{block_number}'.lower()
            rpc_calls[decimals_query_id] = self.state_service.get_function_info(
                address=lp_token, abi=LP_TOKEN_ABI, fn_name="decimals", block_number=block_number)

        return rpc_calls

    def decode_user_info_function(
            self, wallet: str, supplied_data: dict, decoded_data: dict, stake: bool = False,
            block_number: int = "latest"):

        result = {}

        lp_token_info = supplied_data['lp_token_info']
        for lp_token, info in lp_token_info.items():
            try:
                query_id = f'balanceOf_{lp_token}_{wallet}_{block_number}'.lower()
                decimals = decoded_data.get(f'decimals_{lp_token}_{block_number}'.lower())
                amount = decoded_data.get(query_id) / 10 ** decimals

                total_supply = info.get('total_supply')

                result[lp_token] = {
                    "amount": amount,
                    "tokens": {
                        info['token0']: {
                            'idx': 0,
                            'amount': amount * info.get('token0_amount', 0) / total_supply if total_supply else 0
                        },
                        info['token1']: {
                            'idx': 1,
                            'amount': amount * info.get('token1_amount', 0) / total_supply if total_supply else 0
                        }
                    }
                }
            except Exception:
                continue

        return result

    def get_rewards_balance_function_info(self, wallet, supplied_data, block_number: int = "latest"):
        return {}

    def calculate_rewards_balance(
            self, wallet: str, supplied_data: dict, decoded_data: dict, block_number: int = "latest") -> dict:
        return {}

    def get_factory_function_info(self, fn_name, fn_paras, block_number: int = 'latest'):
        factory_addr = self.pool_info.get('factory_address')
        return self.state_service.get_function_info(
            factory_addr, self.factory_abi, fn_name, fn_paras, block_number
        )




import json

UNITROLLER = json.loads("""
[
  {
    "constant": true,
    "inputs": [],
    "name": "pendingAdmin",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function",
    "signature": "0x26782247"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "newPendingAdmin",
        "type": "address"
      }
    ],
    "name": "_setPendingAdmin",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function",
    "signature": "0xb71d1a0c"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "comptrollerImplementation",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function",
    "signature": "0xbb82aa5e"
  },
  {
    "constant": false,
    "inputs": [],
    "name": "_acceptImplementation",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function",
    "signature": "0xc1e80334"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "pendingComptrollerImplementation",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function",
    "signature": "0xdcfbc0c7"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "newPendingImplementation",
        "type": "address"
      }
    ],
    "name": "_setPendingImplementation",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function",
    "signature": "0xe992a041"
  },
  {
    "constant": false,
    "inputs": [],
    "name": "_acceptAdmin",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function",
    "signature": "0xe9c714f2"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "admin",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function",
    "signature": "0xf851a440"
  },
  {
    "inputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "constructor",
    "signature": "constructor"
  },
  {
    "payable": true,
    "stateMutability": "payable",
    "type": "fallback"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldPendingImplementation",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newPendingImplementation",
        "type": "address"
      }
    ],
    "name": "NewPendingImplementation",
    "type": "event",
    "signature": "0xe945ccee5d701fc83f9b8aa8ca94ea4219ec1fcbd4f4cab4f0ea57c5c3e1d815"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldImplementation",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newImplementation",
        "type": "address"
      }
    ],
    "name": "NewImplementation",
    "type": "event",
    "signature": "0xd604de94d45953f9138079ec1b82d533cb2160c906d1076d1f7ed54befbca97a"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldPendingAdmin",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newPendingAdmin",
        "type": "address"
      }
    ],
    "name": "NewPendingAdmin",
    "type": "event",
    "signature": "0xca4f2f25d0898edd99413412fb94012f9e54ec8142f9b093e7720646a95b16a9"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "oldAdmin",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "newAdmin",
        "type": "address"
      }
    ],
    "name": "NewAdmin",
    "type": "event",
    "signature": "0xf9ffabca9c8276e99321725bcb43fb076a6c66a54b7f21c4e8146d8519b417dc"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "name": "error",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "info",
        "type": "uint256"
      },
      {
        "indexed": false,
        "name": "detail",
        "type": "uint256"
      }
    ],
    "name": "Failure",
    "type": "event",
    "signature": "0x45b96fe442630264581b197e84bbada861235052c5a1aadfff9ea4e40a969aa0"
  }
]
""")
        


import json

LIDO_TOP_UP_RECIPIENTS_ABI = json.loads('''
[
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_trustedCaller",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_allowedRecipientsRegistry",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_finance",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_token",
        "type": "address"
      },
      {
        "internalType": "address",
        "name": "_easyTrack",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [],
    "name": "allowedRecipientsRegistry",
    "outputs": [
      {
        "internalType": "contract AllowedRecipientsRegistry",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "address",
        "name": "_creator",
        "type": "address"
      },
      {
        "internalType": "bytes",
        "name": "_evmScriptCallData",
        "type": "bytes"
      }
    ],
    "name": "createEVMScript",
    "outputs": [
      {
        "internalType": "bytes",
        "name": "",
        "type": "bytes"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "bytes",
        "name": "_evmScriptCallData",
        "type": "bytes"
      }
    ],
    "name": "decodeEVMScriptCallData",
    "outputs": [
      {
        "internalType": "address[]",
        "name": "recipients",
        "type": "address[]"
      },
      {
        "internalType": "uint256[]",
        "name": "amounts",
        "type": "uint256[]"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "easyTrack",
    "outputs": [
      {
        "internalType": "contract EasyTrack",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "finance",
    "outputs": [
      {
        "internalType": "contract IFinance",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "token",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "trustedCaller",
    "outputs": [
      {
        "internalType": "address",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  }
]
''')


UNISWAP_V3_NFT_TOKEN_MANGAGER_ABI = [{"inputs": [{"internalType": "address", "name": "_factory", "type": "address"},
                                                 {"internalType": "address", "name": "_WETH9", "type": "address"},
                                                 {"internalType": "address", "name": "_tokenDescriptor_",
                                                  "type": "address"}], "stateMutability": "nonpayable",
                                      "type": "constructor"}, {"anonymous": False, "inputs": [
    {"indexed": True, "internalType": "address", "name": "owner", "type": "address"},
    {"indexed": True, "internalType": "address", "name": "approved", "type": "address"},
    {"indexed": True, "internalType": "uint256", "name": "tokenId", "type": "uint256"}], "name": "Approval",
                                                               "type": "event"}, {"anonymous": False, "inputs": [
    {"indexed": True, "internalType": "address", "name": "owner", "type": "address"},
    {"indexed": True, "internalType": "address", "name": "operator", "type": "address"},
    {"indexed": False, "internalType": "bool", "name": "approved", "type": "bool"}], "name": "ApprovalForAll",
                                                                                  "type": "event"}, {"anonymous": False,
                                                                                                     "inputs": [{
                                                                                                                    "indexed": True,
                                                                                                                    "internalType": "uint256",
                                                                                                                    "name": "tokenId",
                                                                                                                    "type": "uint256"},
                                                                                                                {
                                                                                                                    "indexed": False,
                                                                                                                    "internalType": "address",
                                                                                                                    "name": "recipient",
                                                                                                                    "type": "address"},
                                                                                                                {
                                                                                                                    "indexed": False,
                                                                                                                    "internalType": "uint256",
                                                                                                                    "name": "amount0",
                                                                                                                    "type": "uint256"},
                                                                                                                {
                                                                                                                    "indexed": False,
                                                                                                                    "internalType": "uint256",
                                                                                                                    "name": "amount1",
                                                                                                                    "type": "uint256"}],
                                                                                                     "name": "Collect",
                                                                                                     "type": "event"},
                                     {"anonymous": False, "inputs": [
                                         {"indexed": True, "internalType": "uint256", "name": "tokenId",
                                          "type": "uint256"},
                                         {"indexed": False, "internalType": "uint128", "name": "liquidity",
                                          "type": "uint128"},
                                         {"indexed": False, "internalType": "uint256", "name": "amount0",
                                          "type": "uint256"},
                                         {"indexed": False, "internalType": "uint256", "name": "amount1",
                                          "type": "uint256"}], "name": "DecreaseLiquidity", "type": "event"},
                                     {"anonymous": False, "inputs": [
                                         {"indexed": True, "internalType": "uint256", "name": "tokenId",
                                          "type": "uint256"},
                                         {"indexed": False, "internalType": "uint128", "name": "liquidity",
                                          "type": "uint128"},
                                         {"indexed": False, "internalType": "uint256", "name": "amount0",
                                          "type": "uint256"},
                                         {"indexed": False, "internalType": "uint256", "name": "amount1",
                                          "type": "uint256"}], "name": "IncreaseLiquidity", "type": "event"},
                                     {"anonymous": False, "inputs": [
                                         {"indexed": True, "internalType": "address", "name": "from",
                                          "type": "address"},
                                         {"indexed": True, "internalType": "address", "name": "to", "type": "address"},
                                         {"indexed": True, "internalType": "uint256", "name": "tokenId",
                                          "type": "uint256"}], "name": "Transfer", "type": "event"},
                                     {"inputs": [], "name": "DOMAIN_SEPARATOR",
                                      "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
                                      "stateMutability": "view", "type": "function"},
                                     {"inputs": [], "name": "PERMIT_TYPEHASH",
                                      "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
                                      "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "WETH9",
                                                                                       "outputs": [
                                                                                           {"internalType": "address",
                                                                                            "name": "",
                                                                                            "type": "address"}],
                                                                                       "stateMutability": "view",
                                                                                       "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "to", "type": "address"},
                                                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                         "name": "approve", "outputs": [], "stateMutability": "nonpayable",
                                         "type": "function"},
                                     {"inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
                                      "name": "balanceOf",
                                      "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                      "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "baseURI",
                                                                                       "outputs": [
                                                                                           {"internalType": "string",
                                                                                            "name": "",
                                                                                            "type": "string"}],
                                                                                       "stateMutability": "pure",
                                                                                       "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                      "name": "burn", "outputs": [], "stateMutability": "payable", "type": "function"},
                                     {"inputs": [{"components": [
                                         {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
                                         {"internalType": "address", "name": "recipient", "type": "address"},
                                         {"internalType": "uint128", "name": "amount0Max", "type": "uint128"},
                                         {"internalType": "uint128", "name": "amount1Max", "type": "uint128"}],
                                                  "internalType": "struct INonfungiblePositionManager.CollectParams",
                                                  "name": "params", "type": "tuple"}], "name": "collect",
                                      "outputs": [{"internalType": "uint256", "name": "amount0", "type": "uint256"},
                                                  {"internalType": "uint256", "name": "amount1", "type": "uint256"}],
                                      "stateMutability": "payable", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token0", "type": "address"},
                                                    {"internalType": "address", "name": "token1", "type": "address"},
                                                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                                                    {"internalType": "uint160", "name": "sqrtPriceX96",
                                                     "type": "uint160"}], "name": "createAndInitializePoolIfNecessary",
                                         "outputs": [{"internalType": "address", "name": "pool", "type": "address"}],
                                         "stateMutability": "payable", "type": "function"}, {"inputs": [{"components": [
        {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
        {"internalType": "uint128", "name": "liquidity", "type": "uint128"},
        {"internalType": "uint256", "name": "amount0Min", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Min", "type": "uint256"},
        {"internalType": "uint256", "name": "deadline", "type": "uint256"}],
                                                                                                         "internalType": "struct INonfungiblePositionManager.DecreaseLiquidityParams",
                                                                                                         "name": "params",
                                                                                                         "type": "tuple"}],
                                                                                             "name": "decreaseLiquidity",
                                                                                             "outputs": [{
                                                                                                             "internalType": "uint256",
                                                                                                             "name": "amount0",
                                                                                                             "type": "uint256"},
                                                                                                         {
                                                                                                             "internalType": "uint256",
                                                                                                             "name": "amount1",
                                                                                                             "type": "uint256"}],
                                                                                             "stateMutability": "payable",
                                                                                             "type": "function"},
                                     {"inputs": [], "name": "factory",
                                      "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                      "stateMutability": "view", "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                      "name": "getApproved",
                                      "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                      "stateMutability": "view", "type": "function"}, {"inputs": [{"components": [
        {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
        {"internalType": "uint256", "name": "amount0Desired", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Desired", "type": "uint256"},
        {"internalType": "uint256", "name": "amount0Min", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Min", "type": "uint256"},
        {"internalType": "uint256", "name": "deadline", "type": "uint256"}],
                                                                                                   "internalType": "struct INonfungiblePositionManager.IncreaseLiquidityParams",
                                                                                                   "name": "params",
                                                                                                   "type": "tuple"}],
                                                                                       "name": "increaseLiquidity",
                                                                                       "outputs": [
                                                                                           {"internalType": "uint128",
                                                                                            "name": "liquidity",
                                                                                            "type": "uint128"},
                                                                                           {"internalType": "uint256",
                                                                                            "name": "amount0",
                                                                                            "type": "uint256"},
                                                                                           {"internalType": "uint256",
                                                                                            "name": "amount1",
                                                                                            "type": "uint256"}],
                                                                                       "stateMutability": "payable",
                                                                                       "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "owner", "type": "address"},
                                                    {"internalType": "address", "name": "operator", "type": "address"}],
                                         "name": "isApprovedForAll",
                                         "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
                                         "stateMutability": "view", "type": "function"}, {"inputs": [{"components": [
        {"internalType": "address", "name": "token0", "type": "address"},
        {"internalType": "address", "name": "token1", "type": "address"},
        {"internalType": "uint24", "name": "fee", "type": "uint24"},
        {"internalType": "int24", "name": "tickLower", "type": "int24"},
        {"internalType": "int24", "name": "tickUpper", "type": "int24"},
        {"internalType": "uint256", "name": "amount0Desired", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Desired", "type": "uint256"},
        {"internalType": "uint256", "name": "amount0Min", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Min", "type": "uint256"},
        {"internalType": "address", "name": "recipient", "type": "address"},
        {"internalType": "uint256", "name": "deadline", "type": "uint256"}],
                                                                                                      "internalType": "struct INonfungiblePositionManager.MintParams",
                                                                                                      "name": "params",
                                                                                                      "type": "tuple"}],
                                                                                          "name": "mint", "outputs": [
            {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
            {"internalType": "uint128", "name": "liquidity", "type": "uint128"},
            {"internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"internalType": "uint256", "name": "amount1", "type": "uint256"}], "stateMutability": "payable",
                                                                                          "type": "function"},
                                     {"inputs": [{"internalType": "bytes[]", "name": "data", "type": "bytes[]"}],
                                      "name": "multicall",
                                      "outputs": [{"internalType": "bytes[]", "name": "results", "type": "bytes[]"}],
                                      "stateMutability": "payable", "type": "function"}, {"inputs": [], "name": "name",
                                                                                          "outputs": [
                                                                                              {"internalType": "string",
                                                                                               "name": "",
                                                                                               "type": "string"}],
                                                                                          "stateMutability": "view",
                                                                                          "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                      "name": "ownerOf",
                                      "outputs": [{"internalType": "address", "name": "", "type": "address"}],
                                      "stateMutability": "view", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "spender", "type": "address"},
                                                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
                                                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                                                    {"internalType": "uint8", "name": "v", "type": "uint8"},
                                                    {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                                                    {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
                                         "name": "permit", "outputs": [], "stateMutability": "payable",
                                         "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                      "name": "positions",
                                      "outputs": [{"internalType": "uint96", "name": "nonce", "type": "uint96"},
                                                  {"internalType": "address", "name": "operator", "type": "address"},
                                                  {"internalType": "address", "name": "token0", "type": "address"},
                                                  {"internalType": "address", "name": "token1", "type": "address"},
                                                  {"internalType": "uint24", "name": "fee", "type": "uint24"},
                                                  {"internalType": "int24", "name": "tickLower", "type": "int24"},
                                                  {"internalType": "int24", "name": "tickUpper", "type": "int24"},
                                                  {"internalType": "uint128", "name": "liquidity", "type": "uint128"},
                                                  {"internalType": "uint256", "name": "feeGrowthInside0LastX128",
                                                   "type": "uint256"},
                                                  {"internalType": "uint256", "name": "feeGrowthInside1LastX128",
                                                   "type": "uint256"},
                                                  {"internalType": "uint128", "name": "tokensOwed0", "type": "uint128"},
                                                  {"internalType": "uint128", "name": "tokensOwed1",
                                                   "type": "uint128"}], "stateMutability": "view", "type": "function"},
                                     {"inputs": [], "name": "refundETH", "outputs": [], "stateMutability": "payable",
                                      "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "from", "type": "address"},
                                                    {"internalType": "address", "name": "to", "type": "address"},
                                                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                         "name": "safeTransferFrom", "outputs": [], "stateMutability": "nonpayable",
                                         "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "from", "type": "address"},
                                                    {"internalType": "address", "name": "to", "type": "address"},
                                                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
                                                    {"internalType": "bytes", "name": "_data", "type": "bytes"}],
                                         "name": "safeTransferFrom", "outputs": [], "stateMutability": "nonpayable",
                                         "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                                                    {"internalType": "uint256", "name": "value", "type": "uint256"},
                                                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                                                    {"internalType": "uint8", "name": "v", "type": "uint8"},
                                                    {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                                                    {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
                                         "name": "selfPermit", "outputs": [], "stateMutability": "payable",
                                         "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                                                    {"internalType": "uint256", "name": "nonce", "type": "uint256"},
                                                    {"internalType": "uint256", "name": "expiry", "type": "uint256"},
                                                    {"internalType": "uint8", "name": "v", "type": "uint8"},
                                                    {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                                                    {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
                                         "name": "selfPermitAllowed", "outputs": [], "stateMutability": "payable",
                                         "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                                                    {"internalType": "uint256", "name": "nonce", "type": "uint256"},
                                                    {"internalType": "uint256", "name": "expiry", "type": "uint256"},
                                                    {"internalType": "uint8", "name": "v", "type": "uint8"},
                                                    {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                                                    {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
                                         "name": "selfPermitAllowedIfNecessary", "outputs": [],
                                         "stateMutability": "payable", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                                                    {"internalType": "uint256", "name": "value", "type": "uint256"},
                                                    {"internalType": "uint256", "name": "deadline", "type": "uint256"},
                                                    {"internalType": "uint8", "name": "v", "type": "uint8"},
                                                    {"internalType": "bytes32", "name": "r", "type": "bytes32"},
                                                    {"internalType": "bytes32", "name": "s", "type": "bytes32"}],
                                         "name": "selfPermitIfNecessary", "outputs": [], "stateMutability": "payable",
                                         "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "operator", "type": "address"},
                                                    {"internalType": "bool", "name": "approved", "type": "bool"}],
                                         "name": "setApprovalForAll", "outputs": [], "stateMutability": "nonpayable",
                                         "type": "function"},
                                     {"inputs": [{"internalType": "bytes4", "name": "interfaceId", "type": "bytes4"}],
                                      "name": "supportsInterface",
                                      "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
                                      "stateMutability": "view", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "token", "type": "address"},
                                                    {"internalType": "uint256", "name": "amountMinimum",
                                                     "type": "uint256"},
                                                    {"internalType": "address", "name": "recipient",
                                                     "type": "address"}], "name": "sweepToken", "outputs": [],
                                         "stateMutability": "payable", "type": "function"},
                                     {"inputs": [], "name": "symbol",
                                      "outputs": [{"internalType": "string", "name": "", "type": "string"}],
                                      "stateMutability": "view", "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "index", "type": "uint256"}],
                                      "name": "tokenByIndex",
                                      "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                      "stateMutability": "view", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "owner", "type": "address"},
                                                    {"internalType": "uint256", "name": "index", "type": "uint256"}],
                                         "name": "tokenOfOwnerByIndex",
                                         "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                         "stateMutability": "view", "type": "function"},
                                     {"inputs": [{"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                      "name": "tokenURI",
                                      "outputs": [{"internalType": "string", "name": "", "type": "string"}],
                                      "stateMutability": "view", "type": "function"},
                                     {"inputs": [], "name": "totalSupply",
                                      "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
                                      "stateMutability": "view", "type": "function"}, {
                                         "inputs": [{"internalType": "address", "name": "from", "type": "address"},
                                                    {"internalType": "address", "name": "to", "type": "address"},
                                                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"}],
                                         "name": "transferFrom", "outputs": [], "stateMutability": "nonpayable",
                                         "type": "function"}, {"inputs": [
        {"internalType": "uint256", "name": "amount0Owed", "type": "uint256"},
        {"internalType": "uint256", "name": "amount1Owed", "type": "uint256"},
        {"internalType": "bytes", "name": "data", "type": "bytes"}], "name": "uniswapV3MintCallback", "outputs": [],
                                                               "stateMutability": "nonpayable", "type": "function"}, {
                                         "inputs": [
                                             {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
                                             {"internalType": "address", "name": "recipient", "type": "address"}],
                                         "name": "unwrapWETH9", "outputs": [], "stateMutability": "payable",
                                         "type": "function"}, {"stateMutability": "payable", "type": "receive"}]



import json

LIDO_ARAGON_ACL_ABI = json.loads('''
[
  {
    "constant": true,
    "inputs": [],
    "name": "hasInitialized",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "createBurnedPermission",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "burnPermissionManager",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "grantPermission",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "getPermissionParamsLength",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_paramsHash",
        "type": "bytes32"
      },
      {
        "name": "_who",
        "type": "address"
      },
      {
        "name": "_where",
        "type": "address"
      },
      {
        "name": "_what",
        "type": "bytes32"
      },
      {
        "name": "_how",
        "type": "uint256[]"
      }
    ],
    "name": "evalParams",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "NO_PERMISSION",
    "outputs": [
      {
        "name": "",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_script",
        "type": "bytes"
      }
    ],
    "name": "getEVMScriptExecutor",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "getRecoveryVault",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "CREATE_PERMISSIONS_ROLE",
    "outputs": [
      {
        "name": "",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      },
      {
        "name": "_params",
        "type": "uint256[]"
      }
    ],
    "name": "grantPermissionP",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_who",
        "type": "address"
      },
      {
        "name": "_where",
        "type": "address"
      },
      {
        "name": "_what",
        "type": "bytes32"
      }
    ],
    "name": "hasPermission",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "token",
        "type": "address"
      }
    ],
    "name": "allowRecoverability",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "appId",
    "outputs": [
      {
        "name": "",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "getInitializationBlock",
    "outputs": [
      {
        "name": "",
        "type": "uint256"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "revokePermission",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_token",
        "type": "address"
      }
    ],
    "name": "transferToVault",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      },
      {
        "name": "_index",
        "type": "uint256"
      }
    ],
    "name": "getPermissionParam",
    "outputs": [
      {
        "name": "",
        "type": "uint8"
      },
      {
        "name": "",
        "type": "uint8"
      },
      {
        "name": "",
        "type": "uint240"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_sender",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      },
      {
        "name": "_params",
        "type": "uint256[]"
      }
    ],
    "name": "canPerform",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "getEVMScriptRegistry",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "ANY_ENTITY",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "removePermissionManager",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_newManager",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "setPermissionManager",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      }
    ],
    "name": "getPermissionManager",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_entity",
        "type": "address"
      },
      {
        "name": "_app",
        "type": "address"
      },
      {
        "name": "_role",
        "type": "bytes32"
      },
      {
        "name": "_manager",
        "type": "address"
      }
    ],
    "name": "createPermission",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": false,
    "inputs": [
      {
        "name": "_permissionsCreator",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "payable": false,
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "EMPTY_PARAM_HASH",
    "outputs": [
      {
        "name": "",
        "type": "bytes32"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "kernel",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "isPetrified",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [],
    "name": "BURN_ENTITY",
    "outputs": [
      {
        "name": "",
        "type": "address"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_who",
        "type": "address"
      },
      {
        "name": "_where",
        "type": "address"
      },
      {
        "name": "_what",
        "type": "bytes32"
      },
      {
        "name": "_how",
        "type": "uint256[]"
      }
    ],
    "name": "hasPermission",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "constant": true,
    "inputs": [
      {
        "name": "_who",
        "type": "address"
      },
      {
        "name": "_where",
        "type": "address"
      },
      {
        "name": "_what",
        "type": "bytes32"
      },
      {
        "name": "_how",
        "type": "bytes"
      }
    ],
    "name": "hasPermission",
    "outputs": [
      {
        "name": "",
        "type": "bool"
      }
    ],
    "payable": false,
    "stateMutability": "view",
    "type": "function"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "name": "entity",
        "type": "address"
      },
      {
        "indexed": true,
        "name": "app",
        "type": "address"
      },
      {
        "indexed": true,
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "name": "allowed",
        "type": "bool"
      }
    ],
    "name": "SetPermission",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "name": "entity",
        "type": "address"
      },
      {
        "indexed": true,
        "name": "app",
        "type": "address"
      },
      {
        "indexed": true,
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "name": "paramsHash",
        "type": "bytes32"
      }
    ],
    "name": "SetPermissionParams",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "name": "app",
        "type": "address"
      },
      {
        "indexed": true,
        "name": "role",
        "type": "bytes32"
      },
      {
        "indexed": true,
        "name": "manager",
        "type": "address"
      }
    ],
    "name": "ChangePermissionManager",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "name": "executor",
        "type": "address"
      },
      {
        "indexed": false,
        "name": "script",
        "type": "bytes"
      },
      {
        "indexed": false,
        "name": "input",
        "type": "bytes"
      },
      {
        "indexed": false,
        "name": "returnData",
        "type": "bytes"
      }
    ],
    "name": "ScriptResult",
    "type": "event"
  }
]
''')